/***************************************************************************
 *
 * prefs.c : GNOME Power Preferences
 *
 * Copyright (C) 2005 Richard Hughes, <richard@hughsie.com>
 *
 * Taken in part from:
 * - GNOME Tutorial, example 1
 * http://www.gnome.org/~newren/tutorials/developing-with-gnome/html/apd.html
 *
 * Licensed under the GNU GPL v2.  See COPYING.
 *
 **************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <gtk/gtk.h>
#include <glade/glade.h>
#include <gconf/gconf-client.h>

#include "common.h"
#include "prefs.h"
#include "logger.h"

static GladeXML *all_pref_widgets;

static gboolean system_bus, is_verbose;

static GSList *checks, *combos, *combo_percent, *combo_after, *combo_sleep, *combo_do;

/** Sets/Hides GTK visibility
 *
 *  @param  widgetname		the libglade widget name
 *  @param  set			should widget be visible?
 */
static void
gtk_set_visibility (const char *widgetname, gboolean set)
{
	GtkWidget *widget;
	widget = glade_xml_get_widget (all_pref_widgets, widgetname);
	if (!widget) {
		GPM_ERROR (("widget '%s' failed to load, aborting", widgetname));
		return;
	}

	if (set)
		gtk_widget_show_all (widget);
	else
		gtk_widget_hide_all (widget);
}

/** Sets/Clears GTK Checkbox
 *
 *  @param  widgetname		the libglade widget name
 *  @param  set			should check be ticked?
 */
static void
gtk_set_check (const char *widgetname, gboolean set)
{
	GtkWidget *widget;
	widget = glade_xml_get_widget (all_pref_widgets, widgetname);
	if (!widget) {
		GPM_ERROR (("widget '%s' failed to load, aborting", widgetname));
		return;
	}
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), set);
}

/** Perform the interactive action when a bool gconf key has been changed
 *
 *  @param  key			full gconf key path
 */
static void
gconf_key_action (const char *key)
{
	gboolean value, value2;
	GConfClient *client = gconf_client_get_default ();

	value = gconf_client_get_bool (client, key, NULL);
	if (strcmp (key, "/apps/gnome-power/general/hasUPS") == 0) {
		gtk_set_visibility ("tab_ups", value);
	} else if (strcmp (key, "/apps/gnome-power/general/hasBatteries") == 0) {
		gtk_set_visibility ("tab_battery", value);
		gtk_set_visibility ("frame_computer_batteries", value);
		gtk_set_visibility ("frame_harddrive_batteries", value);
		gtk_set_visibility ("frame_display_batteries", value);
	} else if (strcmp (key, "/apps/gnome-power/general/hasButtons") == 0) {
		gtk_set_visibility ("tab_button", value);
	} else if (strcmp (key, "/apps/gnome-power/general/hasDisplays") == 0) {
		gtk_set_visibility ("tab_display", value);
		value2 = gconf_client_get_bool (client, "/apps/gnome-power/general/hasLCD", NULL);
		gtk_set_visibility ("combo_lcd_mains", value2);
		gtk_set_visibility ("combo_lcd_batteries", value2);
		gtk_set_visibility ("label_lcd_mains", value2);
		gtk_set_visibility ("label_lcd_batteries", value2);
	} else if (strcmp (key, "/apps/gnome-power/general/hasHardDrive") == 0) {
		gtk_set_visibility ("tab_harddrive", value);
	} else if (strcmp (key, "/apps/gnome-power/general/hasWireless") == 0) {
		gtk_set_visibility ("tab_wireless", value);
	} else if (strcmp (key, "/apps/gnome-power/general/hasLCD") == 0) {
		gtk_set_visibility ("combo_lcd_mains", value);
		gtk_set_visibility ("combo_lcd_batteries", value);
		gtk_set_visibility ("label_lcd_mains", value);
		gtk_set_visibility ("label_lcd_batteries", value);

	} else if (strcmp (key, "/apps/gnome-power/icons/showWireless") == 0) {
		gtk_set_check ("check_wireless_display", value);
	} else if (strcmp (key, "/apps/gnome-power/icons/showUPS") == 0) {
		gtk_set_check ("check_ups_display", value);
	} else if (strcmp (key, "/apps/gnome-power/icons/showBatteries") == 0) {
		gtk_set_check ("check_batteries_display", value);
	} else if (strcmp (key, "/apps/gnome-power/icons/showLCD") == 0) {
		gtk_set_check ("check_lcd_display", value);
	} else if (strcmp (key, "/apps/gnome-power/icons/showAcAdapter") == 0) {
		gtk_set_check ("check_ac_adapter_display", value);
	}
}

/** Callback for gconf_key_changed
 *
 */
static void
callback_gconf_key_changed (GConfClient *client, guint cnxn_id, GConfEntry *entry, gpointer user_data)
{
	if (gconf_entry_get_value (entry) == NULL)
		return;

	if (gconf_entry_get_value (entry)->type == GCONF_VALUE_BOOL)
		gconf_key_action (entry->key);
}

/** Callback for combo_changed
 *
 */
static void
callback_combo_changed (GtkWidget *widget, gpointer user_data)
{
	int value, a;
	const char *name;
	GConfClient *client = gconf_client_get_default ();
	gchar *gconfpath;

	value = gtk_combo_box_get_active(GTK_COMBO_BOX (widget));	
	name = gtk_widget_get_name(widget);

	ComboList *data;

	/* find combo name in master combo list */
	int len = g_slist_length (combos);
	gboolean found = FALSE;
	for (a = 0; a < len; a++) {
		data = g_slist_nth_data (combos, a);
		if (strcmp (name, data->widgetname) == 0) {
			found = TRUE;
			break;
		}
	}
	if (!found) {
		GPM_INFO (("Combo '%s' not found, escaping", name));
		return;
	}

	gconfpath = g_strconcat ("/apps/gnome-power/policy/", data->policy, NULL);

	/* get correct actual value from specific combo slist */
	ComboData *data2;
	data2 = g_slist_nth_data (data->dataset, value);
	GPM_INFO (("%s set to logical %i, actual %i", data->policy, value, data2->value));

	gconf_client_set_int (client, gconfpath, data2->value, NULL);
	g_free (gconfpath);
}

/** Callback for button_help
 *
 */
static void
callback_help (GtkWidget *widget, gpointer user_data)
{
	/* for now, show website */
	gnome_url_show ("http://gnome-power.sourceforge.net/", NULL);
}

/** Callback for check_changed
 *
 */
static void
callback_check_changed (GtkWidget *widget, gpointer user_data)
{
	gboolean value;
	const char *widgetname;
	GConfClient *client = gconf_client_get_default ();
	gchar *gconfpath;
	ComboList *data;
	int a;

	widgetname = gtk_widget_get_name(widget);
	value = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget));	

	int len = g_slist_length (checks);
	gboolean found = FALSE;
	for (a = 0; a < len; a++) {
		data = g_slist_nth_data (checks, a);
		if (strcmp (data->widgetname, widgetname) == 0) {
			found = TRUE;
			break;
		}
	}
	if (!found) {
		GPM_WARNING (("Cannot find widget '%s'", widgetname));
		return;
	}

	gconfpath = g_strconcat ("/apps/gnome-power/icons/", data->policy, NULL);

	GPM_INFO (("%s set to logical %i", data->policy, value));

	gconf_client_set_bool (client, gconfpath, value, NULL);
	g_free (gconfpath);
}

/** Description
 *
 *  @param  xxxx		test
 */
static void
check_setup_icons (const char *widgetname, const char *policy)
{
	gchar *policy_path;
	gboolean value;
	GtkWidget *widget;
	GConfClient *client = gconf_client_get_default ();

	policy_path = g_strconcat ("/apps/gnome-power/icons/", policy, NULL);
	value = gconf_client_get_bool (client, policy_path, NULL);
	g_free (policy_path);

	/* do the change before setting up the callback */
	gtk_set_check (widgetname, value);

	widget = glade_xml_get_widget (all_pref_widgets, widgetname);
	if (!widget) {
		GPM_ERROR (("widget '%s' failed to load, aborting", widgetname));
		return;
	}

	g_signal_connect (G_OBJECT (widget), "clicked", G_CALLBACK (callback_check_changed), NULL);
}

/** Adds a combo to it's global SList (for that "type", e.g. percentage)
 *
 *  @param  human_string	User text, e.g. "Hibernate"
 *  @param  value		The real-value, e.g. "5%" value would be 5. 
 *				(Useful as *this* value is stored in gconf not the index)
 */
static void
combo_add (GSList **list, char *human_string, const int value)
{
	ComboData *cdata;
	cdata = g_new (ComboData, 1);
	cdata->human_string = human_string;
	cdata->value = value;
	*list = g_slist_append (*list, cdata);
}

/** Adds a combo to the global SList of combo-boxes
 *
 *  @param  widgetname		the libglade widget name, e.g. combo_batteries_display
 *  @param  policy		the gconf key policy name, e.g. hasBatteries
 *  @param  dataset		what "type" of combo box we are, see function above.
 */
static void
combo_populate (GSList **list, char *widgetname, char *policy, GSList *dataset)
{

	ComboList *cdata;
	cdata = g_new (ComboList, 1);

	cdata->widgetname = widgetname;
	cdata->policy = policy;
	cdata->dataset = dataset;
	*list = g_slist_append (*list, cdata);

	GtkWidget *widget;
	gchar *policy_path;
	gboolean value, found;
	int a;

	widget = glade_xml_get_widget (all_pref_widgets, cdata->widgetname);
	if (!widget) {
		GPM_ERROR (("widget '%s' failed to load, aborting", cdata->widgetname));
		return;
	}

	/* populate human readable string */
	int len = g_slist_length (cdata->dataset);
	for (a = 0; a < len; a++) {
		ComboData *data = g_slist_nth_data (cdata->dataset, a);
		if (data->human_string)
			gtk_combo_box_append_text (GTK_COMBO_BOX (widget), data->human_string);
	}

	/* get value from system gconf */
	GConfClient *client = gconf_client_get_default ();
	policy_path = g_strconcat ("/apps/gnome-power/policy/", cdata->policy, NULL);
	value = gconf_client_get_int (client, policy_path, NULL);
	g_free (policy_path);

	/* this is a human-readable value, e.g. 100 needs to be converted to index 5 for percent */
	found = FALSE;
	for (a = 0; a < len; a++) {
		ComboData *data = g_slist_nth_data (cdata->dataset, a);
		if (value == data->value) {
			found = TRUE;
			break;
		}
	}
	/* check to see if gconf value is sensible */
	if (!found) {
		GPM_WARNING (("GPM policy '%s' got a value '%i' from gconf that it can't match, setting to 0", cdata->widgetname, value));
		a = 0;
	} else
		GPM_INFO (("GPM policy '%s' got a value '%i' from gconf that is ok", cdata->widgetname, value));

	/* set to correct index, and setup callback */
	gtk_combo_box_set_active (GTK_COMBO_BOX (widget), a);
	g_signal_connect (G_OBJECT (widget), "changed", G_CALLBACK (callback_combo_changed), NULL);
}

/** Adds a checkbox to the global SList
 *
 *  @param  widgetname		the libglade widget name
 *  @param  policy		the gconf key policy name, e.g. hasBatteries
 */
static void
check_add (GSList **list, char *widgetname, char *policy)
{

	ComboList *cdata;
	cdata = g_new (ComboList, 1);

	cdata->widgetname = widgetname;
	cdata->policy = policy;
	*list = g_slist_append (*list, cdata);

	check_setup_icons (widgetname, policy);
}

/** Fill the combo boxes with data
 *
 */
static void
populate_combo (void)
{
	GSList **combo_data;
	/* add percentage info */
	combo_data = &combo_percent;
	combo_add (combo_data, "0%", 0);
	combo_add (combo_data, "20%", 20);
	combo_add (combo_data, "40%", 40);
	combo_add (combo_data, "60%", 60);
	combo_add (combo_data, "80%", 80);
	combo_add (combo_data, "100%", 100);

	/* add after info */
	combo_data = &combo_after;
	combo_add (combo_data, "Never", 0);
	combo_add (combo_data, "1 minute", 1);
	combo_add (combo_data, "5 minutes", 5);
	combo_add (combo_data, "10 minutes", 10);
	combo_add (combo_data, "30 minutes", 30);
	combo_add (combo_data, "1 hour", 60);

	/* add suspend info */
	combo_data = &combo_sleep;
	combo_add (combo_data, "Soft suspend", ACTION_SUSPEND);
	/* need to query HAL to see if possible */
	combo_add (combo_data, "Hibernate", ACTION_HIBERNATE);

	/* add do info */
	combo_data = &combo_do;
	combo_add (combo_data, "Do nothing", ACTION_NOTHING);
	combo_add (combo_data, "Send warning", ACTION_WARNING);
	combo_add (combo_data, "Suspend", ACTION_SUSPEND);
/* TODO: if hibernate is supported by the BIOS... can get info from HAL */
	combo_add (combo_data, "Hibernate", ACTION_HIBERNATE);
	combo_add (combo_data, "Shutdown", ACTION_SHUTDOWN);

}

/** Define the references of the combo boxes, i.e. what "type" of combo box should be dealt with
 *
 */
static void
combo_reference (GSList *combos_data)
{

	/* add combo info - i.e. what combo can do what */
	combo_populate (&combos, "combo_sleep_mains", 
			"SleepMainsMinutes", combo_after);
	combo_populate (&combos, "combo_sleep_batteries", 
			"SleepBatteryMinutes", combo_after);
	combo_populate (&combos, "combo_spindown_mains", 
			"SpindownMainsMinutes", combo_after);
	combo_populate (&combos, "combo_spindown_batteries", 
			"SpindownBatteryMinutes", combo_after);
	combo_populate (&combos, "combo_display_mains", 
			"DisplayBlankMainsMinutes", combo_after);
	combo_populate (&combos, "combo_display_batteries", 
			"DisplayBlankBatteryMinutes", combo_after);
	combo_populate (&combos, "combo_ups_fail", 
			"UPSFailAction", combo_do);
	combo_populate (&combos, "combo_ups_low", 
			"UPSLowAction", combo_do);
	combo_populate (&combos, "combo_battery_fail", 
			"BatteryFailAction", combo_do);
	combo_populate (&combos, "combo_battery_low", 
			"BatteryLowAction", combo_do);
	combo_populate (&combos, "combo_wireless_low", 
			"WirelessLowAction", combo_do);
	combo_populate (&combos, "combo_button_power", 
			"ButtonPowerAction", combo_do);
	combo_populate (&combos, "combo_button_sleep", 
			"ButtonSleepAction", combo_do);
	combo_populate (&combos, "combo_button_lid", 
			"ButtonLidAction", combo_do);
	combo_populate (&combos, "combo_sleeptype_mains", 
			"SleepMainsType", combo_sleep);
	combo_populate (&combos, "combo_sleeptype_batteries", 
			"SleepBatteryType", combo_sleep);
	combo_populate (&combos, "combo_lcd_mains", 
			"LCDMainsPercentage", combo_percent);
	combo_populate (&combos, "combo_lcd_batteries", 
			"LCDBatteryPercentage", combo_percent);

}

/** Prints program usage.
 *
 */
static void print_usage (void)
{
	fprintf (stderr, "\nusage : gnome-power-preferences [--system-bus] [--verbose] [--help]\n");
	fprintf (stderr,
		"\n"
		"        --system-bus            Use the gconf system bus\n"
		"        --verbose               Show extra debugging\n"
		"        --help                  Show this information and exit\n"
		"\n");
}

/** Main program
 *
 */
int
main (int argc, char **argv)
{

	gtk_init (&argc, &argv);
	gconf_init (argc, argv, NULL);
	GtkWidget *widget = NULL;
	int a;

	system_bus = FALSE;
	is_verbose = FALSE;
	for (a=1; a < argc; a++) {
		if (strcmp (argv[a], "--verbose") == 0)
			is_verbose = TRUE;
		else if (strcmp (argv[a], "--system-bus") == 0)
			system_bus = TRUE;
		else if (strcmp (argv[a], "--help") == 0) {
			print_usage ();
			return EXIT_SUCCESS;
		}
	}

	if (is_verbose)
		logger_enable ();
	else
		logger_disable ();


	/* load the interface */
	all_pref_widgets = glade_xml_new (GPM_DATA "preferences.glade", NULL, NULL);
	if (!all_pref_widgets) {
		g_print ("glade file failed to load, aborting\n");
		return 0;
	}

	/* Get the GconfClient, tell it we want to monitor /apps/gnome-power */
	GConfClient *client = gconf_client_get_default ();
	gconf_client_add_dir (client, "/apps/gnome-power", GCONF_CLIENT_PRELOAD_NONE, NULL);
	gconf_client_notify_add (client, "/apps/gnome-power", callback_gconf_key_changed, widget, NULL, NULL);

	/* Get the main_window quit */
	widget = glade_xml_get_widget (all_pref_widgets, "window_preferences");
	if (!widget) {
		GPM_ERROR (("Main window failed to load, aborting"));
		return EXIT_FAILURE;
	}
	g_signal_connect (G_OBJECT (widget), "delete_event", G_CALLBACK (gtk_main_quit), NULL);

	/* Get the quit button */
	widget = glade_xml_get_widget (all_pref_widgets, "button_close");
	g_signal_connect (G_OBJECT (widget), "clicked", G_CALLBACK (gtk_main_quit), NULL);

	/* Get the help button */
	widget = glade_xml_get_widget (all_pref_widgets, "button_help");
	g_signal_connect (G_OBJECT (widget), "clicked", G_CALLBACK (callback_help), NULL);

	/* initialise slists */
	combo_percent = NULL;
	combo_after = NULL;
	combo_sleep = NULL;
	combo_do = NULL;
	combos = NULL;
	checks = NULL;

	/* assign the combo boxes their properties */
	populate_combo ();
	combo_reference (combos);

	/* coldplug the values */
	gconf_key_action ("/apps/gnome-power/general/hasUPS");
	gconf_key_action ("/apps/gnome-power/general/hasBatteries");
	gconf_key_action ("/apps/gnome-power/general/hasButtons");
	gconf_key_action ("/apps/gnome-power/general/hasDisplays");
	gconf_key_action ("/apps/gnome-power/general/hasHardDrive");
	gconf_key_action ("/apps/gnome-power/general/hasWireless");
	gconf_key_action ("/apps/gnome-power/general/hasLCD");

	check_add (&checks, "check_ac_adapter_display", 
			"showAcAdapter");
	check_add (&checks, "check_batteries_display", 
			"showBatteries");
	check_add (&checks, "check_ups_display", 
			"showUPS");
	check_add (&checks, "check_lcd_display", 
			"showLCD");
	check_add (&checks, "check_wireless_display", 
			"showWireless");

	/* start the event loop */
	gtk_main ();

	/* free slists */
	g_slist_free (combo_percent);
	g_slist_free (combo_after);
	g_slist_free (combo_sleep);
	g_slist_free (combo_do);
	g_slist_free (combos);
	g_slist_free (checks);

	return 0;
}
